@@ -5,6 +5,11 @@
 //!
 #![deny(missing_debug_implementations)]
 
+#[cfg(not(any(target_os = "macos", target_os = "redox", windows)))]
+#[cfg(feature = "wayland")]
+#[macro_use]
+extern crate dlib;
+
 use std::fmt;
 use std::os::raw;
 
@@ -124,6 +129,8 @@ use self::os::macos as imp;
 use self::os::posix as imp;
 #[cfg(target_os = "redox")]
 use self::os::redox as imp;
+#[cfg(target_arch = "wasm32")]
+use self::os::wasm as imp;
 #[cfg(target_os = "windows")]
 use self::os::windows as imp;
 ///
@@ -804,7 +811,7 @@ impl Window {
     /// Get POSIX menus. Will only return menus on POSIX-like OSes like Linux or BSD
     /// otherwise ```None```
     ///
-    #[cfg(any(target_os = "macos", target_os = "windows"))]
+    #[cfg(any(target_os = "macos", target_os = "windows", target_arch = "wasm32"))]
     pub fn get_posix_menus(&self) -> Option<&Vec<UnixMenu>> {
         None
     }
